/* cap_t.cpp
 *
 * libcap Wrapper for the Microsoft Network Monitor analyzer
 * By Laurent Rabret <laurent.rabret@francetelecom.com>
 * Copyright 2001 FTR&D
 *
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "stdafx.h"

pcap_t_::pcap_t_(DWORD dNum, DWORD slen, int promisc, int to_ms, char *ebuf) :
			deviceNumber(dNum), snaplen(slen), ReadTimeOut(to_ms),
			MinRange(0xFFFF0000 - to_ms), MaxRange(0xFFFFFFFF - to_ms)
{
		// a MAX_SIZE_OF_BUFFER bytes buffer
		datafield = new unsigned char[MAX_SIZE_OF_BUFFER];
		if (!datafield)
			throw -1;
		pBuffer = pBufferLast = pRecycler = NULL;
		pIRTC = NULL;
		fieldbegin = fieldend = 0;
		ZeroMemory(szError, sizeof(szError));
		hBlobWorking = NULL;
		lostPacket	= 0;


		if (InitializeLivePCap(promisc, ebuf) == FALSE) {
			// The destructor won't be called so, we must clean up the instance here
			delete datafield;
			throw -1;
		}
		InitializeCriticalSection(&m_CS);
}


pcap_t_::~pcap_t_()
{
	if (pIRTC) {
		pIRTC->Stop();
		pIRTC->Disconnect();
		// Standard IUnknown call to release the interface
		pIRTC->Release();
	}

	if (pBuffer) {
		delete pBuffer;
	}
	if (pRecycler) {
		delete pRecycler;
	}
	if (datafield) {
		delete datafield;
	}
	if (hBlobWorking) {
		DestroyBlob(hBlobWorking);
	}
	// Don't delete pBufferLast please !! It's just a shortcut
	DeleteCriticalSection(&m_CS);
}

BOOL pcap_t_::AddNewFrame(PUCHAR cbuf, unsigned cSize /*off wire*/ , unsigned len /* on wire*/, long time_sec, long time_usec)
{
	lBuffer* pTemp;
	if (pRecycler == NULL) {
		// There 's no unused "frame handler". So, we must create one
		try {
			pTemp = new lBuffer(cbuf, cSize, len, time_sec, time_usec, fieldbegin, fieldend, datafield);
		}
		catch (...) {
			return FALSE;
		}
		if (!pTemp) {
			return FALSE;
		}
	}
	else {
		// There 's at least one unused "frame handler". Let's take it!!
		EnterCriticalSection(&m_CS);
		pTemp= pRecycler;
		pRecycler = pRecycler->next;
		LeaveCriticalSection(&m_CS);

		if (pTemp->InitFields(cbuf, cSize, len, time_sec, time_usec, fieldbegin, fieldend, datafield) == FALSE) {
			// Initialisation failed, we must reconnect pTemp to pRecycler
			EnterCriticalSection(&m_CS);
			pTemp->next = pRecycler;
			pRecycler = pTemp;
			LeaveCriticalSection(&m_CS);
			return FALSE;	
		}
	}
	// Now, we can build the chain
	EnterCriticalSection(&m_CS);
	if (!pBuffer) {
		pBuffer = pTemp;
	}
	else {
		pBufferLast->next = pTemp;
	}
	LeaveCriticalSection(&m_CS);
	pBufferLast = pTemp;
	return TRUE;
}

/*
Description
	The FindAnalyzer subroutine find the network analyzer BLOB (Binary Large Object)
	associated to an id.

Parameters
	number NIC id (this ID is only valuable for this library)

Return Values
	A handle on the BLOB if found, NULL otherwise
*/

HBLOB pcap_t_::FindAnalyzer(DWORD number)
{
	HBLOB hBlobRet = NULL;

	if (number == 0) return NULL;
	
	HBLOB		FilterBlob  = NULL;
	PBLOB_TABLE pBTable		= NULL;

	CreateBlob(&FilterBlob);
	if (FilterBlob == NULL) {
		return NULL;
	}

	if (NMERR_SUCCESS != (SetBoolInBlob( FilterBlob,
			                                OWNER_NPP,
				                            CATEGORY_CONFIG,
					                        TAG_INTERFACE_REALTIME_CAPTURE,
											TRUE))) {
		DestroyBlob(FilterBlob);
		return NULL;
	}

	if (NMERR_SUCCESS == GetNPPBlobTable( FilterBlob, &pBTable)) {
		// number of valid network tables in pBTable->dwNumBlobs
		for (DWORD i=0; i < pBTable->dwNumBlobs; i++) {
			if (i == (number-1)) {
				hBlobRet = pBTable->hBlobs[i];
			}
			else {
				// destroy unused BLOBS
				DestroyBlob(pBTable->hBlobs[i]);
			}
			pBTable->hBlobs[i] = NULL;
		} // end for
		// Destroy the BLOB table
		DestroyBlob(pBTable);
	}
	DestroyBlob(FilterBlob);
	return hBlobRet;
}


BOOL pcap_t_::InitializeLivePCap(int promisc, char *ebuf)
{
	// The first caracter of the "device" param is a number used to quicly find
	// the correct analyzer
	hBlobWorking = FindAnalyzer(deviceNumber);
	if (!hBlobWorking) {
		strcpy(ebuf, "Can't find analyzer anymore");
		return FALSE;
	}

	if ( CreateNPPInterface(hBlobWorking, IID_IRTC, (void**)&pIRTC) != NMERR_SUCCESS) {
		strcpy(ebuf, "Can't join the low level network analyzer");
		// We mustn't forget to destroy hBlobWorking because not referred 
		return FALSE;
	}

	// To ask the driver to capture snaplen byte / frame max
	NETWORKINFO nInfo;
	DWORD sLen;
	if (GetNetworkInfoFromBlob(hBlobWorking, &nInfo) == NMERR_SUCCESS) {
		sLen = (snaplen > nInfo.MaxFrameSize) ? sLen = RTC_FRAME_SIZE_FULL : snaplen;
	}
	else {
		sLen = snaplen;
	}

	if (SetDwordInBlob(hBlobWorking, OWNER_NPP, CATEGORY_CONFIG, TAG_FRAME_SIZE, sLen) != NMERR_SUCCESS) {
		strcpy(ebuf, "Can't configure snaplen");
		return FALSE;
	}

/*
	// suppressed because not really usefull
	// specify the delay to call the APC
	// must convert from ms to s
	if (SetDwordInBlob(hBlobWorking, OWNER_NPP, CATEGORY_CONFIG, TAG_UPDATE_FREQUENCY, ((ReadTimeOut % 1000) ? 1 : 0) + ReadTimeOut/1000) != NMERR_SUCCESS) {
		strcpy(ebuf, "Can't set callback delay");
		return FALSE;
	}
*/

	// specify whether the card must/mustn't be in promiscuous mode
	if (!promisc) {
		// leave the NIC in non-promiscuous mode
		// by default, NM changes the NIC to promiscuous mode

		// We specify we want all packets and to leave the NIC in non-promiscuous mode
		if (NMERR_SUCCESS != SetNPPEtypeSapFilter(hBlobWorking,
								0, 0, NULL, NULL,
								CAPTUREFILTER_FLAGS_LOCAL_ONLY | CAPTUREFILTER_FLAGS_INCLUDE_ALL,
								NULL)) {
			strcpy(ebuf, "Can't leave the NIC in non-promiscuous mode");
			// NO, we must not retrun FALSE if this call fails
			//return FALSE;
		}

		PCHAR TempString;
		if (NMERR_SUCCESS == GetStringFromBlob(hBlobWorking,
								OWNER_NPP,
								CATEGORY_LOCATION,
								TAG_MACADDRESS,
								(const char**)&TempString)) {

			CHAR MACAddressString[MAC_ADDRESS_SIZE*2+1];
			strcpy(MACAddressString, TempString);
			// now, select our own MAC address only

			ADDRESSTABLE AddTable;
			ZeroMemory(&AddTable, sizeof(AddTable));
			AddTable.nAddressPairs = 2;
			AddTable.AddressPair[0].AddressFlags = ADDRESS_FLAGS_MATCH_DST;
			AddTable.AddressPair[1].AddressFlags = ADDRESS_FLAGS_MATCH_SRC;

			int i=MAC_ADDRESS_SIZE*2-2;

			while (i>=0) {
				AddTable.AddressPair[0].DstAddress.MACAddress[i/2] = (UCHAR)strtoul(MACAddressString + i, 0, 16);
				MACAddressString[i] = 0;
				i-=2;
			}

			memcpy(AddTable.AddressPair[1].SrcAddress.MACAddress, 
				   AddTable.AddressPair[0].DstAddress.MACAddress, MAC_ADDRESS_SIZE);

			if (SetNPPAddressFilterInBlob(hBlobWorking, &AddTable) != NMERR_SUCCESS) {
				strcpy(ebuf, "Can't leave the NIC in non-promiscuous mode");
				return FALSE;
			}
		}
	} // end if (!promisc)
	return TRUE;
}




int	pcap_t_::GetFrames(int cnt, pcap_handler callback, u_char *user, BOOL bWithTimeOut)
{
	int i=0;
	if (pBuffer)
	{
		lBuffer* temp = pBuffer;
		lBuffer* temp2;

		BOOL bTimeOut = ((bWithTimeOut == FALSE) || (ReadTimeOut == 0)) ? FALSE : TRUE;

		DWORD TickLimit = GetTickCount() + ReadTimeOut;
		BOOL bAgain = TRUE;
		do {
			i++;
			callback(user, &temp->pktHead, temp->Buf);
			// Don't need this frame anymore
			fieldbegin = temp->NextBuf;

			EnterCriticalSection(&m_CS);
			temp2 = temp;
			temp = temp->next;
			if (!temp || (cnt == 1)) break;
			LeaveCriticalSection(&m_CS);
			switch (cnt) {
				case 0 : 
					if (bTimeOut) {
						DWORD timenow = GetTickCount();
						// The count can loop...
						if (TickLimit >= 0xFFFF0000) {
							if ((timenow >= TickLimit) || (timenow < MinRange)) {
								bAgain = FALSE;
							}
						}
						else {
							if ((timenow >= TickLimit) && ( timenow < MaxRange))
								bAgain = FALSE;
						}
					} // if (bTimeOut)
					break;
				case -1 :
					// infinite loop
					break;
				default :
					if (--cnt == 0)
						bAgain = FALSE;
			}
		} while (bAgain);

		temp2->next = pRecycler;
		pRecycler	= pBuffer;
		pBuffer		= temp;
		LeaveCriticalSection(&m_CS);
	}
	return i;
}




int	pcap_t_::SetFilter(struct bpf_program *fp)
{
	int		ret = 0;
	HRESULT hRes;
	
	if (!pIRTC) {
		strcpy(szError, "Real Time analyser uninitialized");
		return -1;
	}
	// Let's start the analysis
	// we first set the filter...
	if (pIRTC->Connect(hBlobWorking, NULL, FrameProc,
                                (LPVOID)this, NULL) != NMERR_SUCCESS )
	{
		// the connect failed
		strcpy(szError, "NIC connection failed");
        return -1;
    }
	// ...and run
	if ((hRes = pIRTC->Start()) != NMERR_SUCCESS) {
		switch (hRes) {
			case NMERR_CAPTURE_PAUSED	: strcpy(szError, "Can't start. The capture is in pause mode");
				break;
			case NMERR_CAPTURING		: strcpy(szError, "The capture is already started");
				break;	
			case NMERR_NOT_CONNECTED	: strcpy(szError, "The analyzer is not connected to the network");
				break;
			case NMERR_NOT_REALTIME		: strcpy(szError, "Can't use this analyzer in real time");
				break;
			default :
				strcpy(szError, "Unexpected error");
		}
		ret = -1;
	}
	return ret;
}



int	pcap_t_::DataLink()
{
	DWORD ret;
	DWORD dwMacType;
	if (GetNPPMacTypeAsNumber(hBlobWorking, &dwMacType) == NMERR_SUCCESS){
		switch(dwMacType) {
			case MAC_TYPE_UNKNOWN  : ret = 0; break; // not good be to display something
			case MAC_TYPE_ETHERNET : ret = 1; break;
			case MAC_TYPE_TOKENRING: ret = 6; break; 
			case MAC_TYPE_FDDI	   : ret = 10; break;
			default				   : ret = 0;		 // not good also...
		}
	}
	return ret;
}




int	pcap_t_::Stats(struct pcap_stat *ps)
{
	STATISTICS stat;
	if (!pIRTC) {
		strcpy(szError, "Unexpected field in the pcap struct");
		return -1;
	}
	if  (pIRTC->GetTotalStatistics(&stat,TRUE) == NMERR_SUCCESS)
	{
		ps->ps_recv		= stat.TotalFramesSeen;
		ps->ps_drop		= stat.TotalFramesDropped + lostPacket;
		// Not so sure for ps_ifdrop. Is it dropped by filters ?
		ps->ps_ifdrop	= stat.TotalFramesSeen - stat.TotalFramesFiltered; 
	}
	return 0;
}





/*
Description
	The FrameProc callback function is automatically called by the network analyser driver
	when one of the following condition in encountered:
	1) the buffer is full
	2) the timeslot elapsed and the buffer is not null

Parameters
	Event: descriptor on all captured frames

Return Values
	a network monitor error message. We always return a success status here
*/

DWORD WINAPI  pcap_t_::FrameProc( UPDATE_EVENT Event )
{
	if (!Event.lpUserContext) return NMERR_NO_HANDLES;

	// User parameter is actually a pointer to the pcap struct
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(Event.lpUserContext);

	LPFRAMETABLE lpFrameTable = Event.lpFrameTable;

    // Walk through frames
	// The frame table can wrap the indices
    for ( unsigned i = lpFrameTable->StartIndex; 
				   i != lpFrameTable->EndIndex; 
				   (i == lpFrameTable->FrameTableLength ) ? i=0: i ++ )
	{
        LPFRAME_DESCRIPTOR lpFrameDesc = &lpFrameTable->Frames[i];
		
		// lpFrameDesc->TimeStamp specifies the system time, in microseconds, when the
		// frame was captured. It is a "pure" 64 bits value
		// We need to change reference from January, 1st 1601 to January, 1st 1970 = 11644473600 seconds if I'm right!!
		DWORD time_sec  = (ULONG) (lpFrameDesc->TimeStamp / (__int64)1000000 - 11644473600);
		DWORD time_usec = (ULONG) (lpFrameDesc->TimeStamp % (__int64)1000000);

		if (pcap->AddNewFrame(lpFrameDesc->FramePointer, lpFrameDesc->nBytesAvail, lpFrameDesc->FrameLength, time_sec, time_usec) == FALSE) {
			// we must update the lost packets counter
			pcap->lostPacket += (i > lpFrameTable->EndIndex) ?
				(2 + lpFrameTable->FrameTableLength - i + lpFrameTable->EndIndex) :
				(1 + lpFrameTable->EndIndex - i);
			return NMERR_SUCCESS;
		}
	} // end for
    return NMERR_SUCCESS;
}
