/* pcapexp.cpp
 *
 * libcap Wrapper for the Microsoft Network Monitor analyzer
 * By Laurent Rabret <laurent.rabret@francetelecom.com>
 * Copyright 2001 FTR&D
 *
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "stdafx.h"

/*

  Description
		pcap_open_live()  is  used  to  obtain  a  packet  capture
		descriptor to look at packets on the network.
	   
	Parameters
		device is a string that specifies the network device to open.
		snaplen specifies the maximum number of bytes to capture.
		promisc specifies if the interface is to be put  into  promiscuous
				mode. (Note  that  even  if this parameter is false, the
				interface could well be in promiscuous mode for some other
				reason.)
		to_ms  specifies  the read timeout in milliseconds.
		ebuf is used to return error text and is only set

	Return Value
       when pcap_open_live() fails and returns NULL.
*/

pcap_t *pcap_open_live(char *device, int snaplen, int promisc, int to_ms, char *ebuf) 
{
	// parameters validation
	if (!ebuf) return NULL;

	if (!device || !device[0]) {
		strcpy(ebuf, "Device unspecified");
		return NULL;
	}

	if (snaplen <= 0) {
		strcpy(ebuf, "Invalid buffer size");
		return NULL;
	}

	DWORD DevNum = atoi(device);
	if (atoi == 0) {
		strcpy(ebuf, "Invalid device name");
		return 0;
	}
	// reserve memory to store information
	pcap_t_* pcap;
	try {
		pcap = new pcap_t_(DevNum, snaplen, promisc, to_ms, ebuf);
	}
	catch(...) {
		return NULL;
	}

	if (pcap == NULL) {
		strcpy(ebuf, "Not enough memory");
		return NULL;
	}
	return reinterpret_cast<pcap_t*>(pcap);
}

/*
Description

	pcap_lookupdev()  returns  a  pointer  to a network device
    suitable for use with  pcap_open_live()  and  pcap_lookupnet().

Parameter
	errbuf is used to return error text and is only set

Return Value
	If there is an error, NULL is returned and errbuf
    is filled in with with an appropriate error message.
*/

char *pcap_lookupdev(char *errbuf)
{
	static CHAR driver[4096];
	PCHAR pRet = driver;
	
	HBLOB		FilterBlob  = NULL;
	PBLOB_TABLE pBTable		= NULL;

	try {
		CreateBlob(&FilterBlob);
		if (FilterBlob == NULL) {
			strcpy(errbuf, "Lack of system resources");
			return NULL;
		}

		if (NMERR_SUCCESS != (SetBoolInBlob( FilterBlob,
				                                OWNER_NPP,
					                            CATEGORY_CONFIG,
						                        TAG_INTERFACE_REALTIME_CAPTURE,
												TRUE))) {
			DestroyBlob(FilterBlob);
			strcpy(errbuf, "Can't find suitable network interface");
			return NULL;
		}

		if (NMERR_SUCCESS == GetNPPBlobTable( FilterBlob, &pBTable))
		{
			driver[0] = '\0';
			int base = 0;

			// number of valid network tables in pBTable->dwNumBlobs
			for (DWORD i=0; i < pBTable->dwNumBlobs; i++) {

				PCHAR TempString;
				
				// 1st, we retreive MAC type (Ethernet, Token Ring etc...
				// Note that a dialup PPP connection will have an Ethernet MAC type
				// due to the way WAN NDIS works.
				// We MUST get this parameter
				if (NMERR_SUCCESS == GetStringFromBlob(pBTable->hBlobs[i],
												OWNER_NPP,
												CATEGORY_NETWORKINFO,
												TAG_MACTYPE,
												(const char**)&TempString)) {
					wsprintf(driver+base, "%d_%s", i+1, TempString);
					base += strlen(driver+base);

					
					// Now, let's find if this blob is for RAS connections
					DWORD dwFlag;
					if (NMERR_SUCCESS == GetDwordFromBlob(pBTable->hBlobs[i],
													OWNER_NPP,
													CATEGORY_NETWORKINFO,
													TAG_FLAGS,
													&dwFlag)) {
						if (dwFlag & NETWORKINFO_FLAGS_RAS) {
							strcpy(driver+base, "_DIALUP");
							base += strlen(driver+base);
						}
					}

					// At last, we get the MAC address
					if (NMERR_SUCCESS == GetStringFromBlob(pBTable->hBlobs[i],
													OWNER_NPP,
													CATEGORY_LOCATION,
													TAG_MACADDRESS,
													(const char**)&TempString)) {
						wsprintf(driver+base, "_%s", TempString);
						base += strlen(driver+base);
					}
					
				}
				base += 1+strlen(driver+base); // +1 to indicate next Network Card
				// Don't need this blob anymore
				DestroyBlob(pBTable->hBlobs[i]);
				pBTable->hBlobs[i] = NULL;
			} // end for
			// must complete the string with a double '\0'
			*(driver+base) = '\0';
		}
		else {
			strcpy(errbuf, "Can't find network driver to analyze");
			pRet = NULL;
		}
		DestroyBlob(FilterBlob);
		FilterBlob = NULL;

		// We don't need this BLOB table anymore
		DestroyBlob(pBTable);
		pBTable = NULL;

	}
	catch(...)
	{
		strcpy(errbuf, "Exception caught");
		pRet = NULL;
		if (FilterBlob) {
			DestroyBlob(FilterBlob);
		}
		if (pBTable) {
			DestroyBlob(pBTable);
		}
	}
	return pRet;
}

/*
Description

	pcap_dispatch() is used to collect  and  process  packets.
    
Parameters
	cnt		specifies  the  maximum  number of packets to process
			before returning. A cnt of -1 processes  all  the  packets
			received  in  one buffer. A cnt of 0 processes all packets
			until an error occurs (or EOF is reached).
	callback specifies  a  routine  to  be  called  with three arguments:
			a u_char pointer which is passed in from pcap_dispatch(),
			a pointer  to  the  pcap_pkthdr  struct  (which  precede the
				actual network headers and data), 
			a u_char pointer to the  packet  data.
			
Return Value
	The number of packets read is returned.	Zero is returned when EOF
	is reached in a ``savefile.'' A return  of  -1  indicates an error
	in which case pcap_perror() or pcap_geterr() may be used to display
	the error text.
*/

int pcap_dispatch(pcap_t *p, int cnt, pcap_handler callback, u_char *user) 
{
	if (!p) {
		return -1;
	}
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	return pcap->GetFrames(cnt, callback, user, TRUE);
}

/*
Description
	pcap_loop() is similar to pcap_dispatch() except it  keeps
    reading  packets  until  cnt  packets  are processed or an
    error occurs.  A negative cnt causes pcap_loop()  to  loop
    forever (or at least until an error occurs).
*/
int pcap_loop(pcap_t *p, int cnt, pcap_handler callback, u_char *user)
{
	if (!p) {
		NULL;
	}
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	return  (pcap->GetFrames(cnt, callback, user, FALSE) == -1) ? -1 : 0;
}

/*
Description
	pcap_compile() is used to compile the string str into a
	filter program.  program is a  pointer  to  a  bpf_program
    struct  and is filled in by pcap_compile().  optimize con-
    trols whether optimization on the resulting code  is  per-
    formed. netmask specifies the netmask of the local net.
*/

int pcap_compile(pcap_t *p, struct bpf_program *fp, char *str, int optimize, bpf_u_int32 netmask) 
{
	// We don't support this function yet -> no "hardware" and/or low level filtering performed
	return 0;
}

/*
Description
	pcap_setfilter()  is used to specify a filter program.  fp
    is a pointer to an array of  bpf_program  struct,  usually
    the result of a call to pcap_compile(). -1 is returned on
    failure; 0 is returned on success.
*/

int pcap_setfilter(pcap_t *p, struct bpf_program *fp) 
{
	if (!p) return -1;
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	return pcap->SetFilter(fp);
}

/*
Description
	pcap_datalink()   returns   the   link  layer  type,  e.g.
       DLT_EN10MB.
	NB : returned values can be found in the "libpcap.c" file
*/
int pcap_datalink(pcap_t *p) 
{
	if (!p) return -1;
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	return pcap->DataLink();
}

/*
Description
		pcap_snapshot() returns the snapshot length specified when
		pcap_open_live was called.
*/
int pcap_snapshot(pcap_t *p) 
{
	if (!p) return -1;
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	return pcap->GetSnapShot();
}
 
/*
Description
		int pcap_stats()  returns  0  and  fills  in  a  pcap_stat
		struct.  The  values  represent packet statistics from the
		start of the run to the time of the call. If there  is  an
		error  or  the  under lying packet capture doesn't support
		packet statistics, -1 is returned and the error  text  can
		be obtained with pcap_perror() or pcap_geterr().
*/
int pcap_stats(pcap_t *p, struct pcap_stat *ps) 
{
	if (!p) return -1;
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	return pcap->Stats(ps);
}
 
/*
Description
		pcap_lookupnet() is used to determine the network number
		and  mask associated with the network device device. Both
		netp and maskp are bpf_u_int32 pointers. A return of -1
		indicates an error in which case errbuf is filled in with
		with an appropriate error message.
*/
int pcap_lookupnet(char *device, bpf_u_int32 *netp, bpf_u_int32 *maskp, char *errbuf) 
{
	strcpy(errbuf, "Function not implemented");
	return -1;
}

/*
Description
	pcap_perror() prints the text of the last pcap library
	error on stderr, prefixed by prefix
*/
void pcap_perror(pcap_t *p, char *prefix) 
{
	if (!p) return;
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	cerr<<prefix<<pcap->GetStringError();
}

/*
Description
	pcap_geterr()  returns  the  error  text pertaining to the
    last pcap library error.
*/ 

char *pcap_geterr(pcap_t *p) 
{
	if (!p) {
		return NULL;
	}
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	return pcap->GetStringError();
}

/*
Description
		pcap_close()  closes  the  files  associated  with  p  and
		deallocates resources.
*/

void pcap_close(pcap_t *p) 
{
	if (!p) return;
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	delete pcap;
}

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////   UNSUPPORTED FUNCTIONS HERE   ////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

/*
Description
		pcap_is_swapped() returns true if the current ``savefile''
       uses a different byte order than the current system.
*/
int pcap_is_swapped(pcap_t *p) 
{
	// not supported
	return false;
}
 

/*
Description
*/
int pcap_major_version(pcap_t *p) 
{
	// not supported so version is set to 0.0
	return 0;
}
 
/*
Description
*/
int pcap_minor_version(pcap_t *p) 
{
	// not supported so version is set to 0.0
	return 0;
}

/*
Description
*/
void pcap_dump_close(pcap_dumper_t *p) 
{
	// not supported
}

/*
Description
*/
FILE *pcap_file(pcap_t *p) 
{
	if (!p) {
		return NULL;
	}
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	strcpy(pcap->GetStringError(), "Function not supported");
	return NULL;
}

/*
Description
*/
pcap_t *pcap_open_offline(char *fname, char *ebuf) 
{
	strcpy(ebuf, "Function not supported");
	return NULL;
}

/*
Description
*/
pcap_dumper_t *pcap_dump_open(pcap_t *p, char *fname) 
{
	if (!p) {
		NULL;
	}
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	strcpy(pcap->GetStringError(), "Function not supported");
	return NULL;
}

/*
Description
*/
void pcap_dump(u_char *user, struct pcap_pkthdr *h, u_char *sp)
{
}

/*
Description
*/
const u_char* pcap_next(pcap_t *p, struct pcap_pkthdr *) 
{
	if (!p) {
		NULL;
	}
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	strcpy(pcap->GetStringError(), "Function not supported");
	return NULL;
}

/*
Description
		pcap_fileno()  returns  the  file descriptor number of the
       ``savefile.''
*/
int pcap_fileno(pcap_t *p) 
{
	if (!p) {
		return NULL;
	}
	pcap_t_* pcap = reinterpret_cast<pcap_t_*>(p);
	strcpy(pcap->GetStringError(), "Function not supported");
	return -1;
}

/* 
Description
	pcap_strerror() is  provided  in  case  strerror(1)  isn't
    available.
*/
char *pcap_strerror(int error) 
{
	return "Must return system message here!!!";
}
